DISCLOSURE_SYSTEM_NT16 = 'NT16'
DISCLOSURE_SYSTEM_NT17 = 'NT17'
DISCLOSURE_SYSTEM_NT18 = 'NT18'
DISCLOSURE_SYSTEM_NT19 = 'NT19'
DISCLOSURE_SYSTEM_NL_INLINE_2024 = 'NL-INLINE-2024'
DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER = 'NL-INLINE-2024-GAAP-OTHER'

ALL_NL_INLINE_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_2024,
    DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER,
]

NL_INLINE_GAAP_IFRS_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_2024,
]

NL_INLINE_GAAP_OTHER_DISCLOSURE_SYSTEMS = [
    DISCLOSURE_SYSTEM_NL_INLINE_2024_GAAP_OTHER,
]
