ARL_2022_PREVIEW = "arl-2022-preview"
ARL_2024_PREVIEW = "arl-2024-preview"
ARL_2024_MULTI_TARGET_PREVIEW = "arl-2024-multi-target-preview"

STAND_ALONE_DISCLOSURE_SYSTEMS = [
    ARL_2022_PREVIEW,
    ARL_2024_PREVIEW
]

MULTI_TARGET_DISCLOSURE_SYSTEMS = [
    ARL_2024_MULTI_TARGET_PREVIEW
]
